# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# frozen_string_literal: true

require 'hoss/config/options'
require 'hoss/config/duration'
require 'hoss/config/bytes'
require 'hoss/config/regexp_list'
require 'hoss/config/wildcard_pattern_list'

module Hoss
  # @api private
  class Config
    extend Options

    DEPRECATED_OPTIONS = %i[].freeze

    # rubocop:disable Metrics/LineLength, Layout/ExtraSpacing
    option :config_file,                       type: :string, default: 'config/hoss.yml'
    option :api_host,                          type: :url,    default: ENV['HOSS_API_HOST'] || 'https://app.hoss.com'
    option :api_key,                           type: :string, default: ENV['HOSS_API_KEY']
    option :ingress_host,                      type: :string, default: ENV['HOSS_INGRESS_HOST'] || 'https://ingress.hoss.com/v1'
    option :max_queue_size,                    type: :int,    default: 3000
    option :batch_size,                        type: :int,    default: ENV['HOSS_BATCH_SIZE'] || 512000
    option :max_event_retries,                 type: :int,    default: 100
    option :remote_config_fetch_interval,      type: :int,    default: 300
    option :agentConfig,                       type: :dict
    option :debug,                             type: :boolean, default: ENV['HOSS_DEBUG']
    option :disable_reporting,                 type: :boolean, default: ENV['HOSS_DISABLE_REPORTING'] == 'true'

    option :central_config,                    type: :bool,   default: true
    option :capture_body,                      type: :string, default: 'off'
    option :capture_headers,                   type: :bool,   default: true
    option :capture_env,                       type: :bool,   default: true
    option :current_user_email_method,         type: :string, default: 'email'
    option :current_user_id_method,            type: :string, default: 'id'
    option :current_user_username_method,      type: :string, default: 'username'
    option :custom_key_filters,                type: :list,   default: [],      converter: RegexpList.new
    option :default_labels,                    type: :dict,   default: {}
    option :disable_start_message,             type: :bool,   default: false
    option :disable_instrumentations,          type: :list,   default: %w[]
    option :enabled,                           type: :bool,   default: true
    option :environment,                       type: :string, default: ENV['RAILS_ENV'] || ENV['RACK_ENV']
    option :framework_name,                    type: :string
    option :framework_version,                 type: :string
    option :filter_exception_types,            type: :list,   default: []
    option :global_labels,                     type: :dict
    option :hostname,                          type: :string
    option :http_compression,                  type: :bool,   default: true
    option :instrument,                        type: :bool,   default: true
    option :log_level,                         type: :int,    default: Logger::WARN
    option :log_path,                          type: :string
    option :pool_size,                         type: :int,    default: 1
    option :proxy_address,                     type: :string
    option :proxy_headers,                     type: :dict
    option :proxy_password,                    type: :string
    option :proxy_port,                        type: :int
    option :proxy_username,                    type: :string
    option :recording,                         type: :bool,   default: true
    option :sanitize_field_names,              type: :list,   default: [],      converter: WildcardPatternList.new
    option :server_ca_cert,                    type: :string
    option :service_name,                      type: :string
    option :service_node_name,                 type: :string
    option :service_version,                   type: :string
    option :span_frames_min_duration,          type: :float,  default: '5ms',   converter: Duration.new(default_unit: 'ms')
    option :stack_trace_limit,                 type: :int,    default: 999_999
    option :verify_server_cert,                type: :bool,   default: true

    # rubocop:enable Metrics/LineLength, Layout/ExtraSpacing
    def initialize(options = {})
      @options = load_schema

      if options.is_a? String
        assign(api_key: options)
      else
        assign(options)
      end

      # Pick out config_file specifically as we need it now to load it,
      # but still need the other env vars to have precedence
      env = load_env
      if (env_config_file = env.delete(:config_file))
        self.config_file = env_config_file
      end

      assign(load_config_file)
      assign(env)

      if self.debug == true || self.debug == 'true'
        assign(
          log_level: Logger::DEBUG,
          debug: true,
          log_path: '-'
        )
      end

      unless self.disable_reporting.nil?
        self.disable_reporting = self.disable_reporting == 'true' || self.disable_reporting == true
      end

      yield self if block_given?

      self.logger ||= build_logger

      @__view_paths ||= []
      @__root_path ||= Dir.pwd
    end

    attr_accessor :__view_paths, :__root_path
    attr_accessor :logger

    attr_reader :options

    def assign(update)
      return unless update
      update.each { |key, value| send(:"#{key}=", value) }
    end

    def available_instrumentations
      %w[
        http
        net_http
        faraday
      ]
    end

    def enabled_instrumentations
      available_instrumentations - disable_instrumentations
    end

    def method_missing(name, *args)
      return super unless DEPRECATED_OPTIONS.include?(name)
      warn "The option `#{name}' has been removed."
    end

    def replace_options(new_options)
      return if new_options.nil? || new_options.empty?
      options_copy = @options.dup
      new_options.each do |key, value|
        options_copy.fetch(key.to_sym).set(value)
      end
      @options = options_copy
    end

    def app=(app)
      case app_type?(app)
      when :sinatra
        set_sinatra(app)
      when :rails
        set_rails(app)
      else
        self.service_name = 'ruby'
      end
    end

    def use_ssl?
      api_host.start_with?('https')
    end

    def span_frames_min_duration?
      span_frames_min_duration != 0
    end

    def span_frames_min_duration=(value)
      super
      @span_frames_min_duration_us = nil
    end

    def span_frames_min_duration_us
      @span_frames_min_duration_us ||= span_frames_min_duration * 1_000_000
    end

    def ssl_context
      return unless use_ssl?

      @ssl_context ||=
        OpenSSL::SSL::SSLContext.new.tap do |context|
          if server_ca_cert
            context.ca_file = server_ca_cert
          else
            context.cert_store =
              OpenSSL::X509::Store.new.tap(&:set_default_paths)
          end

          context.verify_mode =
            if verify_server_cert
              OpenSSL::SSL::VERIFY_PEER
            else
              OpenSSL::SSL::VERIFY_NONE
            end
        end
    end

    def inspect
      super.split.first + '>'
    end

    # Deprecations

    def default_tags=(value)
      warn '[DEPRECATED] The option default_tags has been renamed to ' \
        'default_labels.'
      self.default_labels = value
    end

    def custom_key_filters=(value)
      unless value == self.class.schema[:custom_key_filters][:default]
        warn '[DEPRECATED] The option custom_key_filters is being removed. ' \
          'See sanitize_field_names for an alternative.'
      end

      set(:custom_key_filters, value)
    end

    def disabled_instrumentations
      disable_instrumentations
    end

    def active
      enabled
    end
    alias active? active

    def disabled_instrumentations=(value)
      warn '[DEPRECATED] The option disabled_instrumentations has been ' \
        'renamed to disable_instrumentations to align with other agents.'
      self.disable_instrumentations = value
    end

    def active=(value)
      warn '[DEPRECATED] The option active has been renamed to enabled ' \
        'to align with other agents and with the remote config.'
      self.enabled = value
    end

    private

    def load_config_file
      filename = File.join(@__root_path || Dir.pwd, config_file)
      return unless File.exist?(filename)
      read = File.read(filename)
      evaled = ERB.new(read).result
      YAML.safe_load(evaled)
    end

    def load_env
      @options.values.each_with_object({}) do |option, opts|
        next unless (value = ENV[option.env_key])
        opts[option.key] = value
      end
    end

    def build_logger
      Logger.new(log_path == '-' ? STDOUT : log_path).tap do |logger|
        logger.level = log_level
      end
    end

    def app_type?(app)
      if defined?(::Rails::Application) && app.is_a?(::Rails::Application)
        return :rails
      end

      if app.is_a?(Class) && app.superclass.to_s == 'Sinatra::Base'
        return :sinatra
      end

      nil
    end

    def set_sinatra(app)
      self.service_name = format_name(service_name || app.to_s)
      self.framework_name = framework_name || 'Sinatra'
      self.framework_version = framework_version || ::Sinatra::VERSION
      self.__root_path = Dir.pwd
    end

    def set_rails(app)
      self.service_name ||= format_name(service_name || rails_app_name(app))
      self.framework_name ||= 'Ruby on Rails'
      self.framework_version ||= ::Rails::VERSION::STRING
      self.logger ||= ::Rails.logger

      self.__root_path = ::Rails.root.to_s
      self.__view_paths = app.config.paths['app/views'].existent
    end

    def rails_app_name(app)
      if ::Rails::VERSION::MAJOR >= 6
        app.class.module_parent_name
      else
        app.class.parent_name
      end
    end

    def format_name(str)
      str&.gsub('::', '_')
    end
  end
end
