# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# frozen_string_literal: true

# require 'hoss/span/context'

module Hoss
  # @api private
  class Event
    extend Forwardable
    include ChildDurations::Methods

    DEFAULT_TYPE = 'custom'

    class Request
      extend Forwardable
      def initialize
        @headers = {}
      end
      attr_accessor(
        :method,
        :url,
        :headers,
        :body,
        :received_at
      )
    end

    class Response
      extend Forwardable
      def initialize
        @headers = {}
      end
      attr_accessor(
        :status_code,
        :headers,
        :body,
        :received_at
      )
    end

    class Error
      extend Forwardable
      ConnectionTimeout = 'CONNECTION_TIMEOUT'
      RequestTimeout = 'REQUEST_TIMEOUT'
      ConnectionError = 'CONNECTION_ERROR'
      def initialize(type)
        @type = type
      end
      attr_accessor(
        :type,
        :received_at
      )
    end

    # rubocop:disable Metrics/ParameterLists
    def initialize
      @id = SecureRandom.uuid
      @request = Request.new
      # @response = Response.new
      @filtered = false
      @retries = 0
    end
    # rubocop:enable Metrics/ParameterLists

    attr_accessor(
      :request,
      :response,
      :error,
      :filtered,
      :retries,
    )

    attr_reader(
      :id
    )

    # life cycle

    def start(clock_start = Util.monotonic_micros)
      @timestamp = Util.micros
      @clock_start = clock_start
      self
    end

    def stop(clock_end = Util.monotonic_micros)
      @duration ||= (clock_end - @clock_start)
      @self_time = @duration - child_durations.duration
      self
    end

    def done(clock_end: Util.monotonic_micros)
      stop clock_end
      self
    end

    def stopped?
      !!duration
    end

    def started?
      !!timestamp
    end

    def running?
      started? && !stopped?
    end

    # relations

    def inspect
      "<Hoss::Event id:#{id.inspect}" \
        '>'
    end
  end
end
