# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# frozen_string_literal: true

require 'hoss/child_durations'
require 'hoss/event'

module Hoss
  # @api private
  class Instrumenter
    EVENT_KEY = :__elastic_instrumenter_events_key

    include Logging

    # @api private
    class Current
      def initialize
        self.events = []
      end

      def events
        Thread.current[EVENT_KEY] ||= []
      end

      def events=(events)
        Thread.current[EVENT_KEY] ||= []
        Thread.current[EVENT_KEY] = events
      end
    end

    def initialize(config, metrics:, stacktrace_builder:, &enqueue)
      @config = config
      @stacktrace_builder = stacktrace_builder
      @enqueue = enqueue
      @metrics = metrics

      @current = Current.new
    end

    attr_reader :stacktrace_builder, :enqueue

    def start
      debug 'Starting instrumenter'
      # We call register! on @subscriber in case the
      # instrumenter was stopped and started again
      @subscriber&.register!
    end

    def stop
      debug 'Stopping instrumenter'

      current_events.pop until current_events.empty?

      @subscriber&.unregister!
    end

    def handle_forking!
      stop
      start
    end

    def subscriber=(subscriber)
      debug 'Registering subscriber'
      @subscriber = subscriber
      @subscriber.register!
    end


    # events

    def current_events
      @current.events
    end

    def current_event
      current_events.last
    end

    def start_event
      event = Event.new
      current_events.push event
      event.start
    end

    def end_event
      return unless (event = current_events.pop)
      event.done
      enqueue.call event
      event
    end
  end
end
