# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# frozen_string_literal: true

module Hoss
  # @api private
  module Spies
    # @api private
    class HTTPSpy
      TYPE = 'ext'
      SUBTYPE = 'http_rb'

      def self.copy_request_body(body)
        case body.source
        when String
          body.source
        when nil
          nil
        else
          ""
        end
      end

      def install
        ::HTTP::Client.class_eval do
          alias perform_without_apm perform

          def perform(req, options)
            result = nil
            error_in_request = false
            begin
              if req.headers['HOSS-SKIP-INSTRUMENTATION'] == 'true'
                return perform_without_apm(req, options)
              end

              Hoss.with_event do |event|
                event.request.headers['host'] = req.uri.host
                req.headers.each {|n,v| event.request.headers[n] = v}
                event.request.method = req.verb.to_s.upcase
                event.request.url = req.uri.to_s
                event.request.body = Hoss::Spies::HTTPSpy::copy_request_body(req.body)
                event.request.received_at = DateTime.now.strftime('%Q').to_i
                begin
                  result = perform_without_apm(req, options)
                rescue HTTP::TimeoutError => e
                  error_in_request = true
                  event.error = Hoss::Event::Error.new(Hoss::Event::Error::ConnectionTimeout)
                  event.error.received_at = DateTime.now.strftime('%Q').to_i
                  raise
                rescue Exception => e
                  error_in_request = true
                  event.error = Hoss::Event::Error.new(Hoss::Event::Error::ConnectionError)
                  event.error.received_at = DateTime.now.strftime('%Q').to_i
                  raise
                end
                if result
                  event.response = Hoss::Event::Response.new
                  event.response.received_at = DateTime.now.strftime('%Q').to_i
                  event.response.status_code = result.code.to_i
                  result.headers.each {|n,v| event.response.headers[n] = v}
                  event.response.body = result.body.to_s
                end
                result
              end
            rescue Exception => e
              raise if error_in_request
              puts format('Hoss Error: %s %s', e.inspect, e.backtrace)
              return result unless result.nil?
              return perform_without_apm(req, options)
            end
          end
        end
      end
    end

    register 'HTTP', 'http', HTTPSpy.new
  end
end
