# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# frozen_string_literal: true

module Hoss
  # @api private
  module Spies
    # @api private
    class NetHTTPSpy
      KEY = :__hoss_net_http_disabled
      # TYPE = 'ext'
      # SUBTYPE = 'net_http'

      class << self
        def disabled=(disabled)
          Thread.current[KEY] = disabled
        end

        def disabled?
          Thread.current[KEY] ||= false
        end

        def disable_in
          self.disabled = true

          begin
            yield
          ensure
            self.disabled = false
          end
        end
      end

      # rubocop:disable Metrics/CyclomaticComplexity
      def install
        Net::HTTP.class_eval do
          alias request_without_apm request

          def request(req, body = nil, &block)
            result = nil
            error_in_request = false
            begin
              if req['HOSS-SKIP-INSTRUMENTATION'] == 'true' || Hoss::Spies::NetHTTPSpy.disabled?
                return request_without_apm(req, body, &block)
              end

              host = req['host']&.split(':')&.first || address
              method = req.method.to_s.upcase
              path, query = req.path.split('?')

              url = use_ssl? ? +'https://' : +'http://'
              url << host
              url << ":#{port}" if port
              url << path
              url << "?#{query}" if query
              uri = URI(url)

              Hoss.with_event do |event|
                # Record request
                event.request.headers['host'] = uri.hostname
                req.each_header {|n,v| event.request.headers[n] = v}
                event.request.method = method
                event.request.url = uri.to_s
                event.request.body = req.body
                event.request.received_at = DateTime.now.strftime('%Q').to_i

                begin
                  result = request_without_apm(req, body, &block)
                rescue
                  error_in_request = true
                  raise
                end

                if result
                  event.response = Hoss::Event::Response.new
                  event.response.received_at = DateTime.now.strftime('%Q').to_i
                  event.response.status_code = result.code.to_i
                  result.each_header {|n,v| event.response.headers[n] = v}
                  event.response.body = result.body
                end

                result
              end
            rescue Exception => e
              raise if error_in_request
              puts format('Hoss Error: %s %s', e.inspect, e.backtrace)
              return result unless result.nil?
              return request_without_apm(req, body, &block)
            end
          end
        end
      end
      # rubocop:enable Metrics/CyclomaticComplexity
    end

    register 'Net::HTTP', 'net/http', NetHTTPSpy.new
  end
end
