# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# frozen_string_literal: true

module Hoss
  module Transport
    # @api private
    module Serializers
      # @api private
      class UnrecognizedResource < InternalError; end

      # @api private
      class Serializer
        def initialize(config)
          @config = config
        end

        attr_reader :config

        private

        def ms(micros)
          micros.to_f / 1_000
        end

        def keyword_field(value)
          Util.truncate(value)
        end

        def keyword_object(hash)
          return unless hash

          hash.tap do |h|
            h.each { |k, v| hash[k] = keyword_field(v) }
          end
        end

        def mixed_object(hash)
          return unless hash

          hash.tap do |h|
            h.each do |k, v|
              hash[k] = v.is_a?(String) ? keyword_field(v) : v
            end
          end
        end
      end

      # @api private
      class Container
        def initialize(config)
          # @transaction = Serializers::TransactionSerializer.new(config)
          # @span = Serializers::SpanSerializer.new(config)
          @report = Serializers::ReportSerializer.new(config)
          @event = Serializers::EventSerializer.new(config)
          @error = Serializers::ErrorSerializer.new(config)
          @metadata = Serializers::MetadataSerializer.new(config)
          # @metricset = Serializers::MetricsetSerializer.new(config)
        end

        attr_reader :report, :event, :error, :metadata, :metricset
        def serialize(resource)
          case resource
          # when Transaction
          #   transaction.build(resource)
          # when Span
          #   span.build(resource)
          when Report
            report.build(resource)
          when Event
            event.build(resource)
          when Error
            error.build(resource)
          # when Metricset
          #   metricset.build(resource)
          when Metadata
            metadata.build(resource)
          else
            raise UnrecognizedResource, resource.inspect
          end
        end
      end

      def self.new(config)
        Container.new(config)
      end
    end
  end
end

require 'hoss/transport/serializers/context_serializer'
# require 'hoss/transport/serializers/transaction_serializer'
# require 'hoss/transport/serializers/span_serializer'
require 'hoss/transport/serializers/report_serializer'
require 'hoss/transport/serializers/event_serializer'
require 'hoss/transport/serializers/error_serializer'
# require 'hoss/transport/serializers/metricset_serializer'
require 'hoss/transport/serializers/metadata_serializer'
